package hu.kozadat.oaiaggregator.presentation;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.converter.HttpMessageConversionException;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.http.converter.HttpMessageNotWritableException;
import org.springframework.http.converter.xml.AbstractXmlHttpMessageConverter;
import org.springframework.util.ClassUtils;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

import javax.xml.bind.*;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamSource;
import java.io.IOException;
import java.io.StringReader;
import java.util.Set;
import java.util.TreeSet;

/**
 * Created by tnagy on 2015.10.16..
 */
public class OAIPMHAwareJaxb2HttpMessageConverter extends AbstractXmlHttpMessageConverter<Object> implements InitializingBean {

    protected static final String NAMESPACE_PREFIX_MAPPER = "com.sun.xml.bind.namespacePrefixMapper";
    protected static final String JAXB_SCHEMA_LOCATION_OAI_PMH_V2 = "http://www.openarchives.org/OAI/2.0/ http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd";
    protected static final String JAXB_SCHEMA_LOCATION_KOZADAT_ALTALANOS = "http://schema.einfoszab.hu/kozadat/kozadat_altalanos/ http://schema.einfoszab.hu/kozadat/kozadat_altalanos/kozadat_altalanos.xsd";
    protected static final String JAXB_SCHEMA_LOCATION_KOZADAT_KOZLONY = "http://schema.einfoszab.hu/kozadat/kozadat_kozlony/ http://schema.einfoszab.hu/kozadat/kozadat_kozlony/kozadat_kozlony.xsd";
    protected static final String JAXB_SCHEMA_LOCATION_KOZADAT_JOGSZABALYTERVEZET = "http://schema.einfoszab.hu/kozadat/kozadat_jogszabalytervezet/ http://schema.einfoszab.hu/kozadat/kozadat_jogszabalytervezet/kozadat_jogszabalytervezet.xsd";
    protected static final String JAXB_SCHEMA_LOCATION_KOZADAT_OGY_TVJAVASLAT = "http://schema.einfoszab.hu/kozadat/kozadat_OGY_tvjavaslat/ http://schema.einfoszab.hu/kozadat/kozadat_OGY_tvjavaslat/kozadat_OGY_tvjavaslat.xsd";
    protected static final String JAXB_SCHEMA_LOCATION_KOZADAT_ADATBAZIS = "http://schema.einfoszab.hu/kozadat/kozadat_adatbazis/ http://schema.einfoszab.hu/kozadat/kozadat_adatbazis/kozadat_adatbazis.xsd";
    protected static final String JAXB_SCHEMA_LOCATION_KOZADAT_ARCHIV = "http://schema.einfoszab.hu/kozadat/kozadat_archiv/ http://schema.einfoszab.hu/kozadat/kozadat_archiv/kozadat_archiv.xsd";

    protected static final String[] JAXB_SCHEMA_LOCATIONS = new String[] {
            JAXB_SCHEMA_LOCATION_OAI_PMH_V2,
            JAXB_SCHEMA_LOCATION_KOZADAT_ALTALANOS,
            JAXB_SCHEMA_LOCATION_KOZADAT_KOZLONY,
            JAXB_SCHEMA_LOCATION_KOZADAT_JOGSZABALYTERVEZET,
            JAXB_SCHEMA_LOCATION_KOZADAT_OGY_TVJAVASLAT,
            JAXB_SCHEMA_LOCATION_KOZADAT_ADATBAZIS,
            JAXB_SCHEMA_LOCATION_KOZADAT_ARCHIV
    };

    protected static final String OAI_AGGREGATOR_RESPONSE_CONTEXT_PATH = "hu.kozadat.oaiaggregator.service.oaipmh";
    protected static final String[] KOZADAT_CONTEXT_PATH_ARRAY = new String[]{
            "org.purl.dc.elements._1",
            "hu.einfoszab.schema.kozadat.einfoszab_base",
            "hu.einfoszab.schema.kozadat.kozadat_adatbazis",
            "hu.einfoszab.schema.kozadat.kozadat_altalanos",
            "hu.einfoszab.schema.kozadat.kozadat_archiv",
            "hu.einfoszab.schema.kozadat.kozadat_jogszabalytervezet",
            "hu.einfoszab.schema.kozadat.kozadat_kozlony",
            "hu.einfoszab.schema.kozadat.kozadat_ogy_tvjavaslat"
    };
    private static final Logger LOGGER = LoggerFactory.getLogger(OAIPMHAwareJaxb2HttpMessageConverter.class);
    private static final EntityResolver NO_OP_ENTITY_RESOLVER = new EntityResolver() {
        @Override
        public InputSource resolveEntity(String publicId, String systemId) {
            return new InputSource(new StringReader(""));
        }
    };
    private volatile JAXBContext kozadatJaxbContext;
    private Set<String> supportedPackages = new TreeSet<>();
    private String schemaLocations;

    @Override
    protected boolean supports(Class<?> clazz) {
        return supportedPackages.contains(clazz.getPackage().getName());
    }

    @Override
    protected Object readFromSource(Class<?> clazz, HttpHeaders headers, Source source) throws IOException {
        try {
            source = processSource(source);
            Unmarshaller unmarshaller = createUnmarshaller(clazz);
            if (clazz.isAnnotationPresent(XmlRootElement.class)) {
                return unmarshaller.unmarshal(source);
            } else {
                JAXBElement<?> jaxbElement = unmarshaller.unmarshal(source, clazz);
                return jaxbElement.getValue();
            }
        } catch (NullPointerException ex) {
            throw new HttpMessageNotReadableException("NPE while unmarshalling. This can happen on JDK 1.6 due to the presence of DTD declarations, which are disabled.", ex);
        } catch (UnmarshalException ex) {
            throw new HttpMessageNotReadableException("Could not unmarshal to [" + clazz + "]: " + ex.getMessage(), ex);
        } catch (JAXBException ex) {
            throw new HttpMessageConversionException("Could not instantiate JAXBContext: " + ex.getMessage(), ex);
        }
    }

    private Unmarshaller createUnmarshaller(Class<?> clazz) {
        try {
            Unmarshaller unmarshaller = kozadatJaxbContext.createUnmarshaller();
            customizeUnmarshaller(unmarshaller);
            return unmarshaller;
        } catch (JAXBException ex) {
            throw new HttpMessageConversionException("Could not create Unmarshaller for class [" + clazz + "]: " + ex.getMessage(), ex);
        }
    }

    private void customizeUnmarshaller(Unmarshaller unmarshaller) {

    }

    private StringBuilder buildContextPath() {
        StringBuilder contextPath = new StringBuilder();
        contextPath.append(OAI_AGGREGATOR_RESPONSE_CONTEXT_PATH);
        for (String kozadatContextPath : KOZADAT_CONTEXT_PATH_ARRAY) {
            contextPath.append(":");
            contextPath.append(kozadatContextPath);
        }
        return contextPath;
    }

    protected Source processSource(Source source) {
        if (source instanceof StreamSource) {
            StreamSource streamSource = (StreamSource) source;
            InputSource inputSource = new InputSource(streamSource.getInputStream());
            try {
                XMLReader xmlReader = XMLReaderFactory.createXMLReader();
                xmlReader.setFeature("http://apache.org/xml/features/disallow-doctype-decl", true);
                String featureName = "http://xml.org/sax/features/external-general-entities";
                xmlReader.setFeature(featureName, false);
                xmlReader.setEntityResolver(NO_OP_ENTITY_RESOLVER);
                return new SAXSource(xmlReader, inputSource);
            } catch (SAXException ex) {
                logger.warn("Processing of external entities could not be disabled", ex);
                return source;
            }
        } else {
            return source;
        }
    }

    @Override
    protected void writeToResult(Object o, HttpHeaders headers, Result result) throws IOException {
        try {
            Class<?> clazz = ClassUtils.getUserClass(o);
            Marshaller marshaller = createMarshaller(clazz);
            setCharset(headers.getContentType(), marshaller);
            marshaller.marshal(o, result);
        } catch (MarshalException ex) {
            throw new HttpMessageNotWritableException("Could not marshal [" + o + "]: " + ex.getMessage(), ex);
        } catch (JAXBException ex) {
            throw new HttpMessageConversionException("Could not instantiate JAXBContext: " + ex.getMessage(), ex);
        }
    }

    private Marshaller createMarshaller(Class<?> clazz) {
        try {
            Marshaller marshaller = kozadatJaxbContext.createMarshaller();
            customizeMarshaller(marshaller);
            return marshaller;
        } catch (JAXBException ex) {
            throw new HttpMessageConversionException("Could not create Marshaller for class [" + clazz + "]: " + ex.getMessage(), ex);
        }
    }

    private void customizeMarshaller(Marshaller marshaller) throws PropertyException {
        marshaller.setProperty(NAMESPACE_PREFIX_MAPPER, new KozadatNamespacePrefixMapper());
        marshaller.setProperty(Marshaller.JAXB_SCHEMA_LOCATION, schemaLocations);
        marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
        marshaller.setProperty(Marshaller.JAXB_ENCODING, "UTF-8");

    }

    private void setCharset(MediaType contentType, Marshaller marshaller) throws PropertyException {
        if (contentType != null && contentType.getCharSet() != null) {
            marshaller.setProperty(Marshaller.JAXB_ENCODING, contentType.getCharSet().name());
        }
    }


    @Override
    public void afterPropertiesSet() throws Exception {
        StringBuilder contextPath = buildContextPath();
        kozadatJaxbContext = JAXBContext.newInstance(contextPath.toString());

        supportedPackages.add(OAI_AGGREGATOR_RESPONSE_CONTEXT_PATH);
        for (String contextPathElement : KOZADAT_CONTEXT_PATH_ARRAY) {
            supportedPackages.add(contextPathElement);
        }

        StringBuilder schemaLocationsBuilder = new StringBuilder();
        for (int i = 0; i < JAXB_SCHEMA_LOCATIONS.length; i++) {
            String schemaLocation = JAXB_SCHEMA_LOCATIONS[i];
            schemaLocationsBuilder.append(schemaLocation);
            if (i < JAXB_SCHEMA_LOCATIONS.length - 1) {
                schemaLocationsBuilder.append("\n");
            }
        }

        schemaLocations = schemaLocationsBuilder.toString();
    }
}
