package hu.kozadat.oaiaggregator.service;

import hu.kozadat.oaiaggregator.domain.Record;
import hu.kozadat.oaiaggregator.domain.Schema;
import hu.kozadat.oaiaggregator.persistence.KozadatDAO;
import hu.kozadat.oaiaggregator.service.oaipmh.*;
import hu.kozadat.oaicentral.xml.oai_pmh_v2_0.*;
import org.joda.time.DateTime;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.oxm.jaxb.Jaxb2Marshaller;
import org.springframework.stereotype.Service;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.transform.stream.StreamSource;
import java.io.ByteArrayInputStream;
import java.math.BigInteger;
import java.util.Collections;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

/**
 * Created by tnagy on 2015.10.14..
 */
@Service
public class OAIPMHServiceImpl implements OAIPMHService {

    private static final Logger LOGGER = LoggerFactory.getLogger(OAIPMHServiceImpl.class);
    private static final String DATETIME_PATTERN_YYYYMMDDHHMMSS = "yyyyMMddHHmmss";

    @Autowired
    private ErrorFactory errorFactory;
    @Autowired
    private RequestValidator validator;
    @Autowired
    private ResumptionTokenManager tokenManager;
    @Autowired
    private KozadatDAO kozadatDAO;

    @Autowired
    private Jaxb2Marshaller jaxb2Marshaller;

    @Value("${oaipmh.base_url}")
    private String baseUrl;
    @Value("${oaipmh.repository_name}")
    private String repositoryName;
    @Value("${oaipmh.admin_email}")
    private String adminEmail;
    @Value("${oaipmh.maximum_number_of_delivered_identifiers}")
    private int maximumNumberOfDeliveredIds;

    private DateTimeFormatter kozadatDatestampFormatter = DateTimeFormat.forPattern(DATETIME_PATTERN_YYYYMMDDHHMMSS);

    @Override
    public Response process(Request request) {
        Response response = new Response();

        try {

            response.setRequest(buildResponseRequestPart(request));
            response.setResponseDate(getCurrentDate());

            OAIPMHerrorType error = validator.validate(request);

            if (error != null) {

                response.getError().add(error);

            } else {

                processVerbs(request, response);

            }

        } catch (Exception e) {
            LOGGER.error("process(): ERROR: {}", e.getMessage(), e);
            response.getError().add(errorFactory.unexpectedError(e));
        }

        return response;
    }

    private void processVerbs(Request request, OAIPMHtype response) throws DatatypeConfigurationException {

        if (request.isIdentify()) {

            handleIdentify(request, response);

        } else if (request.isListMetadataFormats()) {

            handleListMetadataFormats(request, response);

        } else if (request.isListSets()) {

            handleListSets(request, response);

        } else if (request.isListIdentifiers()) {

            handleListIdentifiers(request, response);

        } else if (request.isListRecords()) {

            handleListRecords(request, response);

        } else if (request.isGetRecord()) {

            handleGetRecord(request, response);

        } else {
            if (LOGGER.isDebugEnabled()) {
                LOGGER.debug("processVerbs(): BAD VERB: {}", request);
            }
            response.getError().add(errorFactory.badVerb());
        }

    }

    /**
     * parameters: resumptionToken (exclusive)
     * errors/exceptions: badArgument, badResumptionToken, noSetHierarchy
     *
     * @param request  the OAI-PMH request
     * @param response the OAI-PMH response
     */
    private void handleListSets(Request request, OAIPMHtype response) {
        response.getError().add(errorFactory.noSetHierarchy());
    }

    /**
     * parameters: identifier (required), metadataPrefix (required)
     * errors/exceptions: badArgument, cannotDisseminateFormat, idDoesNotExist
     *
     * @param request  the OAI-PMH request
     * @param response the OAI-PMH response
     */
    private void handleGetRecord(Request request, OAIPMHtype response) {

        GetRecordType getRecordType = new GetRecordType();

        Record record = findMetadataRecord(request.getIdentifier(), request.getMetadataPrefix());

        if (record == null) {
            response.getError().add(errorFactory.idDoesNotExist(request.getIdentifier()));
        } else {
            RecordType recordType = new RecordType();
            HeaderType headerType = new HeaderType();
            headerType.setIdentifier(record.getOaiId());
            headerType.setDatestamp(JodaTimeUtils.dateToUTCISODateTimeString(kozadatDatestampFormatter.parseDateTime(record.getModified()).toDate()));
            if ("D".equals(record.getState())) {
                headerType.setStatus(StatusType.DELETED);
            }
            recordType.setHeader(headerType);
            MetadataType metadataType = new MetadataType();
            metadataType.setAny(jaxb2Marshaller.unmarshal(new StreamSource(new ByteArrayInputStream(record.getXml()))));
            recordType.setMetadata(metadataType);
            getRecordType.setRecord(recordType);
            response.setGetRecord(getRecordType);
        }

    }

    /**
     * parameters: from (optional), until (optional), metadataPrefix (required), set (optional), resumptionToken (exclusive)
     * errors/exceptions: badArgument, badResumptionToken, cannotDisseminateFormat, noRecordsMatch, noSetHierarchy
     *
     * @param request  the OAI-PMH request
     * @param response the OAI-PMH response
     * @throws DatatypeConfigurationException
     */
    private void handleListRecords(Request request, OAIPMHtype response) throws DatatypeConfigurationException {
        String resumptionToken = request.getResumptionToken();

        String from;
        String until;
        String set;
        String metadataPrefix;

        ListRecordsType listRecordsType = new ListRecordsType();

        List<Long> metadataRecordIds;
        ResumptionToken token = null;

        if (StringUtils.isStringNullOrEmpty(resumptionToken)) {

            set = request.getSet();
            metadataPrefix = request.getMetadataPrefix();
            from = request.getFrom();
            until = request.getUntil();

            Date fromDate = JodaTimeUtils.parseISODateString(from);
            Date untilDate = JodaTimeUtils.parseISODateString(until);
            if (untilDate == null) {
                untilDate = new Date();
            }

            metadataRecordIds = kozadatDAO.findMetadataRecordIds(metadataPrefix, JodaTimeUtils.dateToLocalString(fromDate, DATETIME_PATTERN_YYYYMMDDHHMMSS), JodaTimeUtils.dateToLocalString(untilDate, DATETIME_PATTERN_YYYYMMDDHHMMSS));

        } else {

            token = tokenManager.findByKey(resumptionToken);

            if (token == null) {
                response.getError().add(errorFactory.badResumptionToken());
                set = null;
                metadataPrefix = null;
                metadataRecordIds = null;
            } else {
                set = token.getSet();
                metadataPrefix = token.getMetadataPrefix();
                metadataRecordIds = token.getIdentifiers();
            }
        }

        if (response.getError().isEmpty()) {
            if (!StringUtils.isStringNullOrEmpty(set)) {
                // tnagy sets not supported yet
                response.getError().add(errorFactory.noSetHierarchy());
            } else {

                List<Record> records = null;

                try {

                    if (metadataRecordIds != null && metadataRecordIds.size() > maximumNumberOfDeliveredIds) {
                        if (token == null) {
                            token = tokenManager.create(request);
                            if (LOGGER.isDebugEnabled()) {
                                LOGGER.debug("handleListRecords(): token created={}", token);
                            }

                            token.setIdentifiers(metadataRecordIds);

                        }
                    }

                    List<Long> identifiers = null;
                    // tnagy get next part of identifiers
                    if (token != null) {
                        identifiers = getSubList(token.getIdentifiers(), token.getPage(), token.getMaximumNumberOfDeliveredIdentifiers());
                    }

                    if (identifiers == null || identifiers.isEmpty()) {
                        identifiers = metadataRecordIds;
                    }

                    if (identifiers != null && !identifiers.isEmpty()) {
                        records = kozadatDAO.findMetadataRecords(identifiers);
                    }

                    if (token != null) {

                        // tnagy count elements
                        BigInteger completeListSize = BigInteger.valueOf(token.getIdentifiers().size());

                        // tnagy set page
                        token.setPage(token.getPage() + 1);

                        boolean endOfListReached = false;
                        List<Long> nextSubList = getSubList(token.getIdentifiers(), token.getPage(), token.getMaximumNumberOfDeliveredIdentifiers());
                        if (nextSubList == null || nextSubList.isEmpty()) {
                            endOfListReached = true;
                        }

                        ResumptionTokenType resumptionTokenType = new ResumptionTokenType();

                        if (!endOfListReached) {

                            // tnagy update key
                            token = tokenManager.update(token);

                            // tnagy send back next token
                            resumptionTokenType.setValue(token.getKey());
                            resumptionTokenType.setCursor(token.getCursor());
                            resumptionTokenType.setCompleteListSize(completeListSize);
                            resumptionTokenType.setExpirationDate(getDateAsXMLGregorianCalendar(token.getExpiration()));

                            // tnagy update cursor
                            // tnagy previous cursor value will be shown in current resumptionToken element due to
                            // http://www.openarchives.org/OAI/openarchivesprotocol.html#FlowControl
                            BigInteger cursor = token.getCursor().add(BigInteger.valueOf(records.size()));
                            token.setCursor(cursor);

                        } else {
                            // tnagy must include empty resumptionToken element
                            resumptionTokenType.setCursor(token.getCursor());
                            resumptionTokenType.setCompleteListSize(completeListSize);

                            // tnagy last part of incomplete list -> invalidate token
                            tokenManager.delete(token);
                        }

                        listRecordsType.setResumptionToken(resumptionTokenType);

                    }
                } catch (Exception e) {
                    LOGGER.error(e.getMessage(), e);
                    response.getError().add(errorFactory.cannotDisseminateFormat(metadataPrefix));
                }

                if (records == null || records.isEmpty()) {
                    response.getError().add(errorFactory.noRecordsMatch());
                } else {

                    for (Record record : records) {

                        HeaderType headerType = new HeaderType();
                        headerType.setDatestamp(JodaTimeUtils.dateToUTCISODateTimeString(kozadatDatestampFormatter.parseDateTime(record.getModified()).toDate()));
                        headerType.setIdentifier(record.getOaiId());
                        if ("D".equals(record.getState())) {
                            headerType.setStatus(StatusType.DELETED);
                        }
                        MetadataType metadataType = new MetadataType();

                        metadataType.setAny(jaxb2Marshaller.unmarshal(new StreamSource(new ByteArrayInputStream(record.getXml()))));

                        RecordType recordType = new RecordType();
                        recordType.setHeader(headerType);
                        recordType.setMetadata(metadataType);

                        listRecordsType.getRecord().add(recordType);
                    }

                    response.setListRecords(listRecordsType);
                }

            }
        }
    }

    /**
     * parameters: from (optional), until (optional), metadataPrefix (required), set (optional), resumptionToken (exclusive)
     * errors/exceptions: badArgument, badResumptionToken, cannotDisseminateFormat, noRecordsMatch, noSetHierarchy
     *
     * @param request  the OAI-PMH request
     * @param response the OAI-PMH response
     * @throws DatatypeConfigurationException
     */
    private void handleListIdentifiers(Request request, OAIPMHtype response) throws DatatypeConfigurationException {
        String resumptionToken = request.getResumptionToken();

        String from;
        String until;
        String set;
        String metadataPrefix;

        ListIdentifiersType listIdentifiersType = new ListIdentifiersType();

        List<Long> metadataRecordIds;
        ResumptionToken token = null;

        if (StringUtils.isStringNullOrEmpty(resumptionToken)) {

            set = request.getSet();
            metadataPrefix = request.getMetadataPrefix();
            from = request.getFrom();
            until = request.getUntil();

            Date fromDate = JodaTimeUtils.parseISODateString(from);
            Date untilDate = JodaTimeUtils.parseISODateString(until);
            if (untilDate == null) {
                untilDate = new Date();
            }

            metadataRecordIds = kozadatDAO.findMetadataRecordIds(metadataPrefix, JodaTimeUtils.dateToLocalString(fromDate, DATETIME_PATTERN_YYYYMMDDHHMMSS), JodaTimeUtils.dateToLocalString(untilDate, DATETIME_PATTERN_YYYYMMDDHHMMSS));

        } else {

            token = tokenManager.findByKey(resumptionToken);

            if (token == null) {
                response.getError().add(errorFactory.badResumptionToken());
                set = null;
                metadataPrefix = null;
                metadataRecordIds = null;
            } else {
                set = token.getSet();
                metadataPrefix = token.getMetadataPrefix();
                metadataRecordIds = token.getIdentifiers();
            }
        }

        if (response.getError().isEmpty()) {
            if (!StringUtils.isStringNullOrEmpty(set)) {
                // tnagy sets not supported yet
                response.getError().add(errorFactory.noSetHierarchy());
            } else {

                List<Record> metadataRecords = null;

                try {

                    if (metadataRecordIds != null && metadataRecordIds.size() > maximumNumberOfDeliveredIds) {
                        if (token == null) {
                            token = tokenManager.create(request);

                            token.setIdentifiers(metadataRecordIds);

                        }
                    }

                    List<Long> identifiers = null;
                    // tnagy get next part of identifiers
                    if (token != null) {
                        identifiers = getSubList(token.getIdentifiers(), token.getPage(), token.getMaximumNumberOfDeliveredIdentifiers());
                    }

                    if (identifiers == null || identifiers.isEmpty()) {
                        identifiers = metadataRecordIds;
                    }

                    if (identifiers != null && !identifiers.isEmpty()) {
                        metadataRecords = kozadatDAO.findMetadataRecords(identifiers);
                    }

                    if (token != null) {

                        // tnagy count elements
                        BigInteger completeListSize = BigInteger.valueOf(token.getIdentifiers().size());

                        // tnagy set page
                        token.setPage(token.getPage() + 1);

                        boolean endOfListReached = false;
                        List<Long> nextSubList = getSubList(token.getIdentifiers(), token.getPage(), token.getMaximumNumberOfDeliveredIdentifiers());
                        if (nextSubList == null || nextSubList.isEmpty()) {
                            endOfListReached = true;
                        }

                        ResumptionTokenType resumptionTokenType = new ResumptionTokenType();

                        if (!endOfListReached) {

                            // tnagy update key
                            token = tokenManager.update(token);

                            // tnagy send back next token
                            resumptionTokenType.setValue(token.getKey());
                            resumptionTokenType.setCursor(token.getCursor());
                            resumptionTokenType.setCompleteListSize(completeListSize);
                            resumptionTokenType.setExpirationDate(getDateAsXMLGregorianCalendar(token.getExpiration()));

                            // tnagy update cursor
                            // tnagy old cursor value will be shown in current resumptionToken element due to
                            // http://www.openarchives.org/OAI/openarchivesprotocol.html#FlowControl
                            BigInteger cursor = token.getCursor().add(BigInteger.valueOf(metadataRecords.size()));
                            token.setCursor(cursor);

                        } else {
                            // tnagy must include empty resumptionToken element
                            resumptionTokenType.setCursor(token.getCursor());
                            resumptionTokenType.setCompleteListSize(completeListSize);

                            // tnagy last part of incomplete list -> invalidate token
                            tokenManager.delete(token);
                        }

                        listIdentifiersType.setResumptionToken(resumptionTokenType);

                    }
                } catch (Exception e) {
                    LOGGER.error(e.getMessage(), e);
                    response.getError().add(errorFactory.cannotDisseminateFormat(metadataPrefix));
                }

                if (metadataRecords == null || metadataRecords.isEmpty()) {
                    response.getError().add(errorFactory.noRecordsMatch());
                } else {

                    for (Record record : metadataRecords) {
                        HeaderType headerType = new HeaderType();
                        headerType.setDatestamp(JodaTimeUtils.dateToUTCISODateTimeString(kozadatDatestampFormatter.parseDateTime(record.getModified()).toDate()));
                        headerType.setIdentifier(record.getOaiId());
                        if ("D".equals(record.getState())) {
                            headerType.setStatus(StatusType.DELETED);
                        }
                        listIdentifiersType.getHeader().add(headerType);
                    }

                    response.setListIdentifiers(listIdentifiersType);
                }

            }
        }
    }

    /**
     * parameters: identifier (optional)
     * errors/exceptions: badArgument, idDoesNotExist, noMetadataFormats
     *
     * @param request  the OAI-PMH request
     * @param response the OAI-PMH response
     */
    private void handleListMetadataFormats(Request request, OAIPMHtype response) {
        ListMetadataFormatsType listMetadataFormatsType = new ListMetadataFormatsType();

        List<Schema> metadataSchemas = null;
        if (StringUtils.isStringNullOrEmpty(request.getIdentifier())) {
            metadataSchemas = kozadatDAO.findMetadataSchemas();
        } else {
            Record metadataRecord = findMetadataRecord(request.getIdentifier(), null);
            if (metadataRecord == null) {
                response.getError().add(errorFactory.idDoesNotExist(request.getIdentifier()));
            } else {
                metadataSchemas = Collections.singletonList(metadataRecord.getSchema());
            }
        }

        if (response.getError().isEmpty()) {
            if (metadataSchemas == null || metadataSchemas.isEmpty()) {
                response.getError().add(errorFactory.noMetadataFormats());
            } else {
                for (Schema metadataSchema : metadataSchemas) {
                    MetadataFormatType metadataFormatType = new MetadataFormatType();

                    metadataFormatType.setMetadataNamespace(metadataSchema.getSchemaId());
                    metadataFormatType.setMetadataPrefix(metadataSchema.getSchemaPrefix());
                    metadataFormatType.setSchema(metadataSchema.getSchemaLocation());

                    listMetadataFormatsType.getMetadataFormat().add(metadataFormatType);

                }
            }
        }

        response.setListMetadataFormats(listMetadataFormatsType);
    }

    /**
     * parameters: none
     * errors/exceptions: badArgument
     *
     * @param request  the OAI-PMH request
     * @param response the OAI-PMH response
     */
    private void handleIdentify(Request request, OAIPMHtype response) {
        GranularityType granularityType = GranularityType.YYYY_MM_DD;

        String erliestDateString = kozadatDAO.getEarliestDateString();

        DateTime earliestDatestamp = kozadatDatestampFormatter.parseDateTime(erliestDateString);
        String earliestDatestampString = null;

        if (GranularityType.YYYY_MM_DD.equals(granularityType)) {
            earliestDatestampString = JodaTimeUtils.dateToUTCISODateString(earliestDatestamp.toDate());
        } else if (GranularityType.YYYY_MM_DD_THH_MM_SS_Z.equals(granularityType)) {
            earliestDatestampString = JodaTimeUtils.dateToUTCISODateTimeString(earliestDatestamp.toDate());
        }

        IdentifyType identifyType = new IdentifyType();

        identifyType.setRepositoryName(repositoryName);
        identifyType.setBaseURL(baseUrl);
        identifyType.setProtocolVersion(PROTOCOL_VERSION);
        identifyType.getAdminEmail().add(adminEmail);
        identifyType.setEarliestDatestamp(earliestDatestampString);
        identifyType.setDeletedRecord(DeletedRecordType.NO);
        identifyType.setGranularity(granularityType);

        response.setIdentify(identifyType);
    }

    private List<Long> getSubList(List<Long> completeList, int page, int pageSize) {
        if (completeList == null || completeList.isEmpty()) {
            return null;
        }
        int fromIndex = (page - 1) * pageSize;
        if (fromIndex < 0) {
            fromIndex = 0;
        }
        if (fromIndex > completeList.size()) {
            fromIndex = completeList.size();
        }
        int toIndex = fromIndex + pageSize;
        if (toIndex > completeList.size()) {
            toIndex = completeList.size();
        }

        return completeList.subList(fromIndex, toIndex);
    }

    private RequestType buildResponseRequestPart(Request request) {
        RequestType requestType = new RequestType();
        requestType.setIdentifier(request.getIdentifier());

        String from = request.getFrom();
        try {
            JodaTimeUtils.parseISODateString(from);
            requestType.setFrom(from);
        } catch (Exception e) {
            LOGGER.warn(e.getMessage() + ": from=" + from);
        }

        requestType.setMetadataPrefix(request.getMetadataPrefix());
        requestType.setResumptionToken(request.getResumptionToken());
        requestType.setSet(request.getSet());

        String until = request.getUntil();
        try {
            JodaTimeUtils.parseISODateString(until);
            requestType.setUntil(until);
        } catch (Exception e) {
            LOGGER.warn(e.getMessage() + ": until=" + until);
        }

        requestType.setValue(request.getRequestURL());

        try {
            VerbType verbType = VerbType.fromValue(request.getVerb());
            requestType.setVerb(verbType);
        } catch (Exception e) {
            LOGGER.warn("buildResponseRequestPart(): {}, verb={}", e.getMessage(), request.getVerb());
        }

        return requestType;
    }

    private XMLGregorianCalendar getCurrentDate() throws DatatypeConfigurationException {
        return getDateAsXMLGregorianCalendar(new Date());
    }

    private XMLGregorianCalendar getDateAsXMLGregorianCalendar(Date date) throws DatatypeConfigurationException {
        GregorianCalendar gregorianCalendar = new GregorianCalendar();
        gregorianCalendar.setTime(date);
        XMLGregorianCalendar xmlGregorianCalendar = DatatypeFactory.newInstance().newXMLGregorianCalendar(gregorianCalendar);
        return xmlGregorianCalendar;
    }

    private Record findMetadataRecord(String identifier, String metadataPrefix) {
        try {
            return kozadatDAO.findMetadataRecord(identifier, metadataPrefix);
        } catch (EmptyResultDataAccessException e) {
            return null;
        }
    }

}
