package hu.kozadat.oaiaggregator.service.oaipmh;

import hu.kozadat.oaiaggregator.service.JodaTimeUtils;
import hu.kozadat.oaicentral.xml.oai_pmh_v2_0.OAIPMHerrorType;
import hu.kozadat.oaicentral.xml.oai_pmh_v2_0.OAIPMHerrorcodeType;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

/**
 * Created by tnagy on 2015.10.14..
 */
@Component
public class RequestValidator {

    private static final String ERROR_MESSAGE_PART_INVALID_PARAMETER = "Invalid Parameter: ";
    private static final String ERROR_MESSAGE_PART_REQUIRED_PARAMETER = "Required Parameter: ";
    private static final String ERROR_MESSAGE_PART_EXCLUSIVE_PARAMETER = "Exclusive Parameter: ";
    private static final String ERROR_MESSAGE_PART_CANNOT_PARSE_UTC_DATE = "Cannot parse UTC date: ";

    @Autowired
    private ErrorFactory errorFactory;

    public OAIPMHerrorType validate(Request oaiPmhRequest) {
        if (oaiPmhRequest.isIdentify()) {
            return validateIdentify(oaiPmhRequest);
        } else if (oaiPmhRequest.isListMetadataFormats()) {
            return validateListMetadataFormats(oaiPmhRequest);
        } else if (oaiPmhRequest.isListSets()) {
            return validateListSets(oaiPmhRequest);
        } else if (oaiPmhRequest.isListIdentifiers()) {
            return validateListIdentifiers(oaiPmhRequest);
        } else if (oaiPmhRequest.isListRecords()) {
            return validateListRecords(oaiPmhRequest);
        } else if (oaiPmhRequest.isGetRecord()) {
            return validateGetRecord(oaiPmhRequest);
        }
        return errorFactory.badVerb();
    }

    protected boolean hasArgument(Request oaiPmhRequest) {
        return oaiPmhRequest.getFrom() != null
                || oaiPmhRequest.getIdentifier() != null
                || oaiPmhRequest.getMetadataPrefix() != null
                || oaiPmhRequest.getSet() != null
                || oaiPmhRequest.getUntil() != null;
    }

    protected OAIPMHerrorType validateIdentify(Request oaiPmhRequest) {
        if (oaiPmhRequest.getFrom() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "from");
        }
        if (oaiPmhRequest.getIdentifier() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "identifier");
        }
        if (oaiPmhRequest.getMetadataPrefix() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "metadataPrefix");
        }
        if (oaiPmhRequest.getResumptionToken() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "resumptionToken");
        }
        if (oaiPmhRequest.getSet() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "set");
        }
        if (oaiPmhRequest.getUntil() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "until");
        }
        return null;
    }

    protected OAIPMHerrorType validateListMetadataFormats(Request oaiPmhRequest) {
        if (oaiPmhRequest.getFrom() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "from");
        }
        if (oaiPmhRequest.getMetadataPrefix() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "metadataPrefix");
        }
        if (oaiPmhRequest.getResumptionToken() != null) {
            errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "resumptionToken");
        }
        if (oaiPmhRequest.getSet() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "set");
        }
        if (oaiPmhRequest.getUntil() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "until");
        }
        return null;
    }

    protected OAIPMHerrorType validateListSets(Request oaiPmhRequest) {
        if (oaiPmhRequest.getResumptionToken() == null) {
            if (oaiPmhRequest.getFrom() != null) {
                return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "from");
            }
            if (oaiPmhRequest.getIdentifier() != null) {
                return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "identifier");
            }
            if (oaiPmhRequest.getMetadataPrefix() != null) {
                return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "metadataPrefix");
            }
            if (oaiPmhRequest.getSet() != null) {
                return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "set");
            }
            if (oaiPmhRequest.getUntil() != null) {
                return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "until");
            }
        } else {
            if (hasArgument(oaiPmhRequest)) {
                return errorFactory.badArgument(ERROR_MESSAGE_PART_EXCLUSIVE_PARAMETER + "resumptionToken");
            }
        }
        return null;
    }

    protected OAIPMHerrorType validateListIdentifiers(Request oaiPmhRequest) {
        if (oaiPmhRequest.getResumptionToken() == null) {
            if (oaiPmhRequest.getIdentifier() != null) {
                return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "identifier");
            }
            if (oaiPmhRequest.getMetadataPrefix() == null) {
                return errorFactory.badArgument(ERROR_MESSAGE_PART_REQUIRED_PARAMETER + "metadataPrefix");
            }
            if (oaiPmhRequest.getFrom() != null) {
                try {
                    JodaTimeUtils.parseISODateString(oaiPmhRequest.getFrom());
                } catch (Exception e) {
                    return errorFactory.badArgument(ERROR_MESSAGE_PART_CANNOT_PARSE_UTC_DATE + "from=" + oaiPmhRequest.getFrom());
                }
            }
            if (oaiPmhRequest.getUntil() != null) {
                try {
                    JodaTimeUtils.parseISODateString(oaiPmhRequest.getUntil());
                } catch (Exception e) {
                    return errorFactory.badArgument(ERROR_MESSAGE_PART_CANNOT_PARSE_UTC_DATE + "until=" + oaiPmhRequest.getUntil());
                }
            }
        } else {
            if (hasArgument(oaiPmhRequest)) {
                return errorFactory.badArgument(ERROR_MESSAGE_PART_EXCLUSIVE_PARAMETER + "resumptionToken");
            }
        }
        return null;
    }

    protected OAIPMHerrorType validateListRecords(Request oaiPmhRequest) {
        OAIPMHerrorType oaiPmhError = new OAIPMHerrorType();
        if (oaiPmhRequest.getResumptionToken() == null) {
            if (oaiPmhRequest.getIdentifier() != null) {
                oaiPmhError.setCode(OAIPMHerrorcodeType.BAD_ARGUMENT);
                oaiPmhError.setValue(ERROR_MESSAGE_PART_INVALID_PARAMETER + "identifier");
                return oaiPmhError;
            }
            if (oaiPmhRequest.getMetadataPrefix() == null) {
                oaiPmhError.setCode(OAIPMHerrorcodeType.BAD_ARGUMENT);
                oaiPmhError.setValue(ERROR_MESSAGE_PART_REQUIRED_PARAMETER + "metadataPrefix");
                return oaiPmhError;
            }
            if (oaiPmhRequest.getFrom() != null) {
                try {
                    JodaTimeUtils.parseISODateString(oaiPmhRequest.getFrom());
                } catch (Exception e) {
                    oaiPmhError.setCode(OAIPMHerrorcodeType.BAD_ARGUMENT);
                    oaiPmhError.setValue(ERROR_MESSAGE_PART_CANNOT_PARSE_UTC_DATE + ": from=" + oaiPmhRequest.getFrom());
                    return oaiPmhError;
                }
            }
            if (oaiPmhRequest.getUntil() != null) {
                try {
                    JodaTimeUtils.parseISODateString(oaiPmhRequest.getUntil());
                } catch (Exception e) {
                    oaiPmhError.setCode(OAIPMHerrorcodeType.BAD_ARGUMENT);
                    oaiPmhError.setValue(ERROR_MESSAGE_PART_CANNOT_PARSE_UTC_DATE + ": until=" + oaiPmhRequest.getUntil());
                    return oaiPmhError;
                }
            }
        } else {
            if (hasArgument(oaiPmhRequest)) {
                oaiPmhError.setCode(OAIPMHerrorcodeType.BAD_ARGUMENT);
                oaiPmhError.setValue(ERROR_MESSAGE_PART_EXCLUSIVE_PARAMETER + "resumptionToken");
                return oaiPmhError;
            }
        }
        return null;
    }

    protected OAIPMHerrorType validateGetRecord(Request oaiPmhRequest) {
        if (oaiPmhRequest.getFrom() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "from");
        }
        if (oaiPmhRequest.getIdentifier() == null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_REQUIRED_PARAMETER + "identifier");
        }
        if (oaiPmhRequest.getMetadataPrefix() == null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_REQUIRED_PARAMETER + "metadataPrefix");
        }
        if (oaiPmhRequest.getResumptionToken() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "resumptionToken");
        }
        if (oaiPmhRequest.getSet() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "set");
        }
        if (oaiPmhRequest.getUntil() != null) {
            return errorFactory.badArgument(ERROR_MESSAGE_PART_INVALID_PARAMETER + "until");
        }
        return null;
    }

}
