package hu.kozadat.oaiaggregator.service.oaipmh;

import hu.kozadat.oaiaggregator.service.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import java.math.BigInteger;
import java.util.*;
import java.util.concurrent.locks.ReentrantReadWriteLock;

/**
 * Created by tnagy on 2015.10.14..
 */
@Component
public class ResumptionTokenManager {

    private static final Logger LOGGER = LoggerFactory.getLogger(ResumptionTokenManager.class);

    private static final long DEFAULT_EXPIRATION_PERIOD = 300000;
    private static final int DEFAULT_MAXIMUM_NUMBER_OF_DELIVERED_IDENTIFIERS = 100;

    private Map<String, ResumptionToken> tokens = new TreeMap<>();
    private volatile ReentrantReadWriteLock lock = new ReentrantReadWriteLock();

    @Value("${oaipmh.resumption_token_expiration_period}")
    private long expirationPeriod = DEFAULT_EXPIRATION_PERIOD;
    @Value("${oaipmh.maximum_number_of_delivered_identifiers}")
    private int maximumNumberOfDeliveredIdentifiers = DEFAULT_MAXIMUM_NUMBER_OF_DELIVERED_IDENTIFIERS;

    public ResumptionToken findByKey(String key) {
        try {
            lock.readLock().lock();

            return tokens.get(key);

        } finally {
            lock.readLock().unlock();
        }
    }

    public void delete(ResumptionToken token) {
        try {
            lock.writeLock().lock();

            tokens.remove(token.getKey());

        } finally {
            lock.writeLock().unlock();
        }
    }

    public ResumptionToken update(ResumptionToken oldToken) {
        return _update(oldToken);
    }

    public ResumptionToken create(Request request) {

        ResumptionToken resumptionToken = new ResumptionToken();
        resumptionToken.setCreated(new Date());
        resumptionToken.setPage(1);
        resumptionToken.setCursor(BigInteger.valueOf(0l));
        resumptionToken.setFrom(request.getFrom());
        resumptionToken.setMaximumNumberOfDeliveredIdentifiers(maximumNumberOfDeliveredIdentifiers);
        resumptionToken.setMetadataPrefix(request.getMetadataPrefix());
        resumptionToken.setSet(request.getSet());
        resumptionToken.setUntil(request.getUntil());

        return _update(resumptionToken);
    }

    public void removeExpired() {
        LOGGER.info("removeExpired(): Managing {} tokens. Removing expired tokens...", tokens.size());
        try {
            lock.writeLock().lock();

            Date now = new Date();

            List<ResumptionToken> tokensToRemove = new LinkedList<>();

            for (ResumptionToken resumptionToken : tokens.values()) {
                if (resumptionToken.getExpiration().before(now)) {
                    if (LOGGER.isDebugEnabled()) {
                        LOGGER.debug("removeExpired(): Expired token found: {}. Token expired={}, now={}", resumptionToken.getKey(), resumptionToken.getExpiration(), now);
                    }
                    tokensToRemove.add(resumptionToken);
                }
            }

            LOGGER.info("removeExpired(): Reference date is {}, found {} token(s) to remove.", now, tokensToRemove.size());

            int removed = 0;
            for (ResumptionToken resumptionToken : tokensToRemove) {
                tokens.remove(resumptionToken.getKey());
                removed++;
            }

            LOGGER.info("removeExpired(): Removing expired tokens finished, a total number of {} token(s) removed.", removed);
        } finally {
            lock.writeLock().unlock();
        }
    }

    protected ResumptionToken _update(ResumptionToken resumptionToken) {
        try {
            lock.writeLock().lock();

            if (!StringUtils.isStringNullOrEmpty(resumptionToken.getKey())) {
                tokens.remove(resumptionToken.getKey());
            }

            resumptionToken.setKey(String.format("%x", resumptionToken.getCreated().getTime()) + "@" + String.format("%x", resumptionToken.getPage()));
            resumptionToken.setExpiration(new Date(resumptionToken.getCreated().getTime() + expirationPeriod));

            tokens.put(resumptionToken.getKey(), resumptionToken);

            return resumptionToken;
        } finally {
            lock.writeLock().unlock();
        }
    }

}
